#include "EmPronoun.h"
#include <pthread.h>
#include <math.h>
#include "Gov3.h"

#define NUMANTES 10
#define MAXDIST 64
#define EXPLPROB 0.1  //for best overall, 0.1 and 1.0/5.5
//#define EXPLPROB 0.8
#define GENCUTOFF 10
#define NUMSENTTYP 3
#define NUMSYNTAXTYP 3
#define NUMANTEWTYP 6
#define NUMPROWTYP 4
#define NUMPROTYP 3
#define NUMANTETYP 4
#define ANTEDISTSZ (NUMSENTTYP*NUMSYNTAXTYP*NUMANTEWTYP*NUMPROWTYP*NUMPROTYP*NUMANTETYP)
#define ANTESENTSEP (NUMSYNTAXTYP*NUMANTEWTYP*NUMPROWTYP*NUMPROTYP*NUMANTETYP)
#define SYNTAXNOSEP (NUMANTEWTYP*NUMPROWTYP*NUMPROTYP*NUMANTETYP)
#define ANTEWORDSEP (NUMPROWTYP*NUMPROTYP*NUMANTETYP)
#define PROTYPSEP   (NUMPROTYP*NUMANTETYP)
#define ANTETYPSEP  (NUMANTETYP)
static pthread_mutex_t reslock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t pggalock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t pngnlock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t pagdlock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t pgghlock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t ppgplock = PTHREAD_MUTEX_INITIALIZER;

vector<stIndex> EmPronoun::Pronouns;
vector<stIndex> EmPronoun::Whwords;
vector<stIndex> EmPronoun::Preps;
vector<EvalData> OPrefnumToData;
vector<EvalData> nOPrefnumToData;
  
int
maxNP(Tree* t)
{
  Tree* p=t->parent;
  if(!p)return 1;
  if(p->htree!=t->htree)return 1;
  stIndex tmp=t->label;
  string tlabs=Tree::gst->toString(tmp);
  if(tlabs!="NP")return 1;
  Tree* r;
  int numNPs=0;
  for(r=p->subtrees;r;r=r->sibling)
    if(Tree::gst->toString(r->label)=="NP")numNPs++;
  if(numNPs<2)return 0;
  else return 1;
}

Tree*
refTree(Tree* t)
{
  if(!t)return NULL;
  if(!t->subtrees) return refTree(t->parent);
  stIndex temp=t->label;
  string tlabs=Tree::gst->toString(t->label);
  if(tlabs=="PRP$")return t;
  else if(tlabs=="NP"&& maxNP(t)) return t;
  else return refTree(t->parent);
}

int
EmPronoun::
anteType(stIndex sti, stIndex wi,StoryInfo& stI)
{
  if(whPronoun(wi)>=0)return 0;
  string stiw=Tree::gst->toString(sti);
  if(stiw=="NNP"||stiw=="NNPS")      return 2;
  else if(stiw=="NN"||stiw=="NNS")return 3;
  else return 1;
}

int
EmPronoun::
syntaxConstraint(AnteInfo ai)
{
  SummaryInfo* asi=ai.ante;
  SummaryInfo* psi=ai.pro;
  if(asi->pretrm==Tree::gst->toIndex("EX")) return 1;
  /* reflexives cannot be coreferent with something not in this sentence*/
  if(whReflexivePronoun(psi->wrd)>=0 && asi->sentNo!=psi->sentNo)return 1;
  if(asi->sentNo!=psi->sentNo)return 0;
  if(asi->wrd==psi->gov&&psi->syntaxNo>NP&&psi->syntaxNo<PASSOBJ) return 1;
  if(asi->gov==psi->gov&& asi->syntaxNo==0&& psi->syntaxNo==1
     &&whReflexivePronoun(psi->wrd)<0)return 1;
  /* more of the constraint that "you"s do not have antes in our corpus */
  if(whPronoun(asi->wrd)&& whPerson(asi->wrd)==SECOND) return 1;
  if(asi->refNo==-2) return 1;//cerr<<"SAW-2"<<endl;
  return 0;
}

stIndex
EmPronoun::
cannonize(stIndex wrd)
{
  int i=whPronoun(wrd);
  if(i<0)return wrd;
  else if(i<4)return Pronouns[0];
  else if(i<8)return Pronouns[4];
  else if(i<11)return Pronouns[8];
  else if(i<16)return Pronouns[11];
  else if(i<20)return Pronouns[16];
  else if(i<23)return Pronouns[20];
  else if(i<27)return Pronouns[23];
  assert(0);
  return 0;
}

void
EmPronoun::
setPronouns()
{
  string prns[27]= {"he","him","his","himself","she","her","hers","herself",
		    "it","its","itself", /*11 */
		    "they","them","theirs","their","themselves", /* 16 */
		    "i","me","my","myself", /* 20 */
		    "you","your","yourself", /* 23 */
		    "we","us","our","ourselves" /* 27 */
                    };
  string preps[21]={"about", "after", "against", "among ", "as",
	"at", "before", "by", "during", "for", "from", "into",
		    "of", "on", "over", "since", "through", "to","under", "until", "with"};
  string wh[4]={"who","that","where","which"};
  int i,j;
  for(i=0;i<27;i++){
    stIndex sti=Tree::gst->toIndex(prns[i]);
    //cerr<<"PRN "<<prns[i]<<" "<<sti<<endl;
    Pronouns.push_back(sti);
  }
  for(i=0;i<4;i++) Whwords.push_back(Tree::gst->toIndex(wh[i]));
  for(i=0;i<21;i++)
    Preps.push_back(Tree::gst->toIndex(preps[i]));
  for(i=0;i<27;i++){
    stIndex pretrm=Tree::gst->toIndex("PRP");
    if(whPossessivePronoun(Pronouns[i])>=0)pretrm=Tree::gst->toIndex("PRP$");
    stIndex proCmb =pretrm+(50*Pronouns[i]);
    PGen& pg=genMap[proCmb];
    if(i<11 || i>=16){
      for(j=0;j<3;j++)pg.prob[j]=0;
      if(i<4)pg.prob[0]=1;
      else if(i<8)pg.prob[1]=1;
      else if(i<11)pg.prob[2]=1;
      if(i>=16){
	pg.prob[0]=.5;
	pg.prob[1]=.5;
      }
    }
  }
  DdTrip ddp;
  ddp.prob=0;
  ddp.pos=0;
  ddp.cnt=0;
  for(i=0;i<ANTEDISTSZ;i++) pAntegDist.push_back(ddp);
  for(i=0;i<ANTEDISTSZ;i+=ANTESENTSEP)pAntegDist[i].prob=1;
  for(i=0;i<3;i++){
    for(j=0;j<2;j++){
      pNumgNum[i][j].first=1;
      pNumgNum[i][j].second=0;
    }
  }
  for(i=0;i<12;i++){
    for(j=0;j<3;j++){
      pPersongPerson[i][j].first=1;
      pPersongPerson[i][j].second=0;
    }
  }
  for(i=0;i<15;i++)subResults[i]=0;
  results[0]=results[1]=0;
  string treestr="classes.txt";
  ifstream ifsc(treestr.c_str());
  //assert(ifsc); //???;
  //gov.readStems();
  //gov.wclassRead(ifsc);
}

int
EmPronoun::
whWh(stIndex sti)
{
  for(int i=0;i<4;i++)if(sti==Whwords[i])return i;
  return -1;
}

void
EmPronoun::
writeProbs()
{
  int i;
  ofstream ofs("probs.txt");
  ofs<< "pAntegDist\n";
  for(i=0;i<ANTEDISTSZ;i++){
    if(i%ANTESENTSEP==0)ofs<<"\n";
    ofs<<"\t"<<pAntegDist[i].prob;
    if(i%ANTEWORDSEP==(ANTEWORDSEP-1))ofs<<"\n";
  }
  ofs<<"\npNumgNum\n";
  for(i=0;i<3;i++)ofs<<"\t"<<pNumgNum[i][0].first<<endl;
  ofs<<"\npPersongPerson\n";
  for(i=0;i<12;i++){
    for(int j=0;j<3;j++)
      ofs<<pPersongPerson[i][j].first<<"\t";
    ofs<<"\n";
  }
  ofs<<"\npGengWord\n";
  GenMap::iterator gmI=genMap.begin();
  for(;gmI!=genMap.end();gmI++){
    PGen& pg=gmI->second;
    double cnt=0;
    for(i=0;i<3;i++)cnt+=pg.pos[i];
    if(cnt<GENCUTOFF)continue;
    stIndex cst=gmI->first;
    int i;
    stIndex wst=cst/50;
    stIndex pst=cst%50;
    if(wst>0)ofs<<Tree::gst->toString(wst)<<"\t";
    else ofs<<".\t";
    ofs<<Tree::gst->toString(pst)<<"\t";
    for(i=0;i<3;i++)      ofs<<pg.prob[i]<<"\t";
    ofs<<"\n";
  }
  ofstream govstr("govestr");
  gov.writeGovs(govstr);
}

void
EmPronoun::
readProbs()
{
  int i,j;
  ifstream ifs("probs.txt");
  string str;
  ifs>> str;
  for(i=0;i<ANTEDISTSZ;i++)
    ifs>>pAntegDist[i].prob;
  ifs >>str;
  for(i=0;i<3;i++){
    ifs>>pNumgNum[i][0].first;
    pNumgNum[i][1].first=1-pNumgNum[i][0].first;
  }

  ifs>>str;
  for(i=0;i<12;i++){
    for(int j=0;j<3;j++)
      ifs>>pPersongPerson[i][j].first;
  }

float vls[72]=
  {0, 0, 0, 1, 0, 0,
   0, 0, 1, 1, 0, 0,
   0, 1, 0, 0, 0, 1,
   0, 1, 1, 1, 0, 0,
   1, 0, 0, 0, 1, 0,
   1, 0, 1, 0, 1, 0,
   1, 1, 0, 0, 0, 1,
   1, 1, 1, 0, 1, 0,
   2, 0, 0, 0, 0, 1,
   2, 0, 1, .1, .1, .8,
   2, 1, 0, .1, .1, .8,
   2, 1, 1, 0, 0, 1};
 for(i=0;i<62;i){
   int a=(int)vls[i++];
   a = (a*2)+(int)vls[i++];
   a = (a*2)+(int)vls[i++];
   for(j=0;j<3;j++){
     //pPersongPerson[a][j].first=vls[i++];
     pPersongPerson[a][j].second=0;
   }
 }
  ifs>>str;
  for(;;){
    ifs >>str;
    if(!ifs)break;
    stIndex wi=0;
    if(str!=".")wi=Tree::gst->toIndex(str);
    //cerr<<str<<" ";
    ifs>>str;
    stIndex pi=Tree::gst->toIndex(str);
    stIndex ci=(wi*50)+pi;
    PGen& pg = genMap[ci];
    for(i=0;i<3;i++) ifs>>pg.prob[i];
    //cerr<<str<<" "<<wi<<" "<<pi<<" "<<genMap[ci].prob[0]<<endl;
  }
}

int
EmPronoun::
whPerson(stIndex lab)
{
  int wp=whPronoun(lab);
  if(wp<0)return THIRD;
  if(wp<16)return THIRD;
  else if(wp<20)return FIRST;
  else if(wp<23)return SECOND;
  else return FIRST;
}

int
EmPronoun::
whPrep(stIndex lab)
{
  int i;
  for(i=0;i<21;i++){
    if(lab==Preps[i])break;
  }
  if(i>=21)return -1;
  else return i;
}

int
EmPronoun::
isPlural(stIndex wrd)
{
  int wp=whPronoun(wrd);
  assert(whPerson(wrd)!=SECOND);
  if(wp<0){
    assert(wp>=0);
  }
  if(wp<11) return 0;
  else if (wp<16)return 1;
  else if(wp<20)return 0;
  return 1;
}

int 
EmPronoun::
proGen(stIndex wrd)
{
  int whp=whPronoun(wrd);
  assert(whp>=0);
  if(whp<4)return 0;
  else if(whp<8)return 1;
  else if(whp<11)return 2;
  return -1;
}

int
EmPronoun::
whPronoun(stIndex lab)
{
  int i;
  for(i=0;i<27;i++){
    if(lab==Pronouns[i])break;
  }
  if(i>=27)return -1;
  else return i;
}

int
EmPronoun::
proType(stIndex wrd)
{
  if(whPossessivePronoun(wrd)>=0)return 2;
  else if(whReflexivePronoun(wrd)>=0)return 1;
  return 0;
}

int
EmPronoun::
whReflexivePronoun(stIndex wrd)
{
  int pnum=whPronoun(wrd);
  if(pnum==3)return 3;
  else if(pnum==7)return 7;
  else if(pnum==10)return 10;
  else if(pnum==15)return 15;
  else if(pnum==19)return 19;
  else if(pnum==22)return 22;
  else if(pnum==26)return 26;
  else return -1;
}

int
EmPronoun::
whPossessivePronoun(stIndex lab)
{
  int pnum=whPronoun(lab);
  if(pnum<2)return -1;
  else if(pnum==2)return 2;
  else if(pnum==6) return 6;
  else if(pnum ==9) return 9;
  else if(pnum==13)return 13;
  else if(pnum==14)return 14;
  else if(pnum==18)return 18;
  else if(pnum==21)return 21;
  else if(pnum==25)return 25;
  return -1;
}

void
EmPronoun::
outputStory(ostream& os,StoryInfo& stI)
{
  if(stI.storyNo%1000==0)cerr<<stI.storyNo<<endl;
  os<<stI.storyNo<<"\t"<<stI.sums.size()<<endl;
  SummaryInfos::iterator siI=stI.sums.begin();
  for(;siI!=stI.sums.end();siI++){
    siI->write(os);
  }
  os.flush();
} 

int
EmPronoun::
bucketDist(int dist)
{
  int bucks[5] = {2,4,8,16,32};
  int ans=0;
  for(;ans<5;ans++)
    if(dist<=bucks[ans])return ans;
  return ans;
}

double
EmPronoun::
pNumGNum(AnteInfo& ai)
{
  stIndex pro=ai.pro->wrd;
  if(whPerson(pro)==SECOND)return 1;
  string preterm=Tree::gst->toString(ai.ante->pretrm);
  stIndex wrd=ai.ante->wrd;
  int plrl=2;
  if(whPronoun(wrd)>=0&& whPerson(wrd)!=SECOND){
    if(isPlural(wrd))plrl=1;
    else plrl=0;
  }
  else if(preterm=="NN" || preterm=="NNP") plrl=0;
  else if(preterm=="NNS" ||preterm=="NNPS") plrl=1;
  int proplrl=isPlural(pro);
  DdPair& ddp=pNumgNum[plrl][proplrl];
  if(ai.prob>=0){
    pthread_mutex_lock(&pngnlock);
    ddp.second+=ai.prob;
    pthread_mutex_unlock(&pngnlock);
    //cerr<<proplrl<<" "<<plrl<<" "<<ddp.second<<endl;
  }
  return ddp.first;
}

int
EmPronoun::
whAnte(Tree* t)
{
  if(whWh(t->label)>=0){
    Tree* p =t->parent;
    if(!p)return -1;
    p=p->parent;
    if(!p)return -1;
    p=p->parent;
    if(!p)return -1;
    string pl=Tree::gst->toString(p->label);
    if(pl!="SBAR")return -1;
    p=p->parent;
    if(!p)return -1;
    if(Tree::gst->toString(p->label)!="NP")return -1;
    int wp=p->htree->wPos();
    //cerr<<wp<<endl;
    return wp;
  }
  return -1;
}

double
EmPronoun::
pPersonGPerson(AnteInfo& ai)
{

  int perA= whPerson(ai.ante->wrd);
  int perP= whPerson(ai.pro->wrd);
  int aQ = ai.ante->quoteNo;
  int pQ = ai.pro->quoteNo;
  perA = (perA*2)+ aQ;
  perA = (perA*2)+pQ;
  assert(perA<12&& perA>=0);
  DdPair& ddp=pPersongPerson[perA][perP];
  double prb=ai.prob;
  if(prb>=0){
    pthread_mutex_lock(&ppgplock);
    ddp.second+=prb;
    pthread_mutex_unlock(&ppgplock);
  }
  return ddp.first;
}

double
EmPronoun::
distributeProb(double ans,int cmb)
{
  if(ans>0)return ans;
  if(stage>ANTE3)return ans;
  int newc;
  double nans;

  newc=cmb-(cmb%PROTYPSEP);
  DdTrip& ddp=pAntegDist[newc];
  nans=ddp.prob;
  if(nans>0)return nans;

  newc=cmb-(cmb%ANTEWORDSEP);
  ddp=pAntegDist[newc];
  nans=ddp.prob;
  if(nans>0)return nans;

  newc=cmb-(cmb%SYNTAXNOSEP);
  ddp=pAntegDist[newc];
  nans=ddp.prob;
  if(nans>0)return nans;

  newc=cmb-(cmb%ANTESENTSEP);
  ddp=pAntegDist[newc];
  nans=ddp.prob;
  return nans;
}

double
EmPronoun::
pAnteGDist(AnteInfo& ai,StoryInfo& stI)
{
  int synNo=ai.ante->syntaxNo;
  if(synNo>=PASSOBJ&&synNo<LASTSYN)synNo=SUBJECT;

  if(synNo==IS)synNo=SUBJECT;
  else if(synNo>3)synNo=2;
  assert(synNo<3);
  int sdist = stI.sentNo-ai.ante->sentNo;
  if(sdist>=3){
    cerr<<sdist<<endl;
    assert(sdist<3);
  }
  int dist = ai.ante->wordNo;
  if(sdist==0)dist=stI.wordNo-dist;
  int cmb =sdist*ANTESENTSEP;
  if(stage>=ANTE2){
    cmb += synNo*SYNTAXNOSEP;
    cmb += bucketDist(dist)*ANTEWORDSEP;
    int d2=bucketDist(stI.wordNo);
    if(d2>3)d2=3;
    cmb+=d2*PROTYPSEP;
  }
  if(stage>=ANTE3){
    cmb+= ANTETYPSEP*proType(ai.pro->wrd);
    int at=anteType(ai.ante->pretrm,ai.ante->wrd,stI);
    cmb+=at;
  }
  
  double prb=ai.prob;
  assert(cmb<ANTEDISTSZ);
  DdTrip& ddp=pAntegDist[cmb];
  //double ans=ddp.prob;
  double ans=distributeProb(ddp.prob,cmb);
  //cerr<<"\t"<<ans<<" "<<ddp.prob<<" "<<cmb<<endl;
  if(ai.prob>=0){
    pthread_mutex_lock(&pagdlock);
    ddp.pos+=ai.prob;
    ddp.cnt++;
    pthread_mutex_unlock(&pagdlock);
  }
  return ans;
}

double
EmPronoun::
pGovGAnte(AnteInfo& ai)
{
  stIndex anteW=0;
  int pos=0;
  if(ai.ante){
    anteW=ai.ante->wrd;
    pos=ai.ante->pretrm;
  }
  if(stage<GOV) return 1;
  if(ai.pro->gov==0)return 1;
    float  pgov=gov.pGov(anteW,ai.pro->gov,ai.pro->syntaxNo,pos);
    float pgovsq=pow(pgov,1.0/6.0);  //for best overall.
    return pgovsq;
}

double
EmPronoun::
computeAnteProb(AnteInfo& ai,StoryInfo& stI)
{
  double pNum=1;
  double pante=1;
  double pgen=1;
  double pper=1;
  if(ai.ante){ //explitives don't use any of this stuff;
    pNum=pNumGNum(ai);
    pante=pAnteGDist(ai,stI);
    pgen=pGenGHead(ai);
    pper=pPersonGPerson(ai);
  }
  double pgov=pGovGAnte(ai);
  double ans=    pNum* pante* pgen* pper*pgov;
  //cerr<<"\t"<<Tree::gst->toString(ai.ante->wrd);
  //cerr<<" "<< ans<<" "<<pgov<<" "
  //<<pNum<<" "<<pante<<" "<<pper<<" "<<pgen<<endl;
  return ans;
}

bool
EmPronoun::
isNPHead(Tree* tr, Tree* wrd)
{
  if(!tr) return false;
  if(whPossessivePronoun(tr->label)>=0)return true;
  if(!tr->subtrees)return isNPHead(tr->parent,wrd);
  if(!tr->subtrees->subtrees) return isNPHead(tr->parent,wrd);
  if(tr->htree->subtrees !=wrd) return false;
  if(Tree::gst->toString(tr->label)=="NP") return true;
  return isNPHead(tr->parent,wrd);
}

void
EmPronoun::
proc(int i,StoryInfo& stI)
{
  SummaryInfos& sums=stI.sums;
  SummaryInfo& si=sums[i];
  stI.wordNo=si.wordNo;
  stI.sentNo=si.sentNo;
  int synno=si.syntaxNo;
  if(synno==3)    return;

  string pret=Tree::gst->toString(si.pretrm);
  /* require both word and part-of-speech to ID as pronoun (e.g. "us") */
  if(whPronoun(si.wrd)<0 || (pret!="PRP" && pret!="PRP$")){
    return;
  }

  AnteInfos anteInfos;
  SummaryInfos::reverse_iterator siI=sums.rbegin(); 
  vector<double> probs;
  int j=i-1;
  for(;j>=0;j--){
    if(synno==3&& !anteInfos.empty())break; //wh words always last np
    SummaryInfo& asi=sums[j];
    if((stI.sentNo-asi.sentNo)>2)break;
    if(asi.syntaxNo==3)continue;
    AnteInfo ai(&si,&asi);
    if(syntaxConstraint(ai))continue;
    ai.prob=computeAnteProb(ai,stI);
    anteInfos.push_back(ai);
  }
  if(ISIT(si.wrd)&&stage>=GOV){
    AnteInfo ai0(&si,NULL);
    float pg=pGovGAnte(ai0);
    float prb=EXPLPROB*pg;
    //cerr<<"\t"<<"EXPL "<<prb<<" "<<pg<<endl;
    ai0.prob=prb;
    anteInfos.push_back(ai0);
  }
  annotate(anteInfos,stI,si);
}

void
EmPronoun::
writeDictionary(ofstream& dictstr)
{
  vector<string>::iterator vsI=Tree::gst->sarray.begin();
  vsI++;
  int i=1;
  for(;vsI!=Tree::gst->sarray.end();vsI++){
    dictstr<<i<<"\t"<<*vsI<<"\n";
    i++;
  }
}

void
EmPronoun::
readDictionary(ifstream& dictstr)
{
  for(;;){
    int i=1;
    int j,k;
    string s;
    dictstr>>j;
    if(!dictstr)break;
    assert(i=j);
    dictstr>>s;
    k=Tree::gst->toIndex(s);
    assert(i=k);
    i++;
  }
}

void
EmPronoun::
summarize(Tree* t,StoryInfo& stI)
{
  Tree* p;
  for(p=t->subtrees;p;p=p->sibling)    summarize(p,stI);
  if(t->subtrees){
    return;
  }
  setquote(t,stI);
  int st=-1;
  st=syntaxType(t);
  int wha=whAnte(t);
  int dq;
  if(wha>=0){
    st=3;
    dq=wha;
    if(wha>=stI.wordNo){
      stI.wordNo++;
      return;
    }
  }
  else dq=stI.doubleQuote<0?0:1;
  int srel;
  if(st<0)st=syntaxType(t);
  stIndex gov=findGov(t,srel);
  if(st==3||isNPHead(t,t)||whPronoun(t->label)>=0){
    SummaryInfo si(stI.storyNo,stI.sentNo,stI.wordNo,st,dq,t);
    if(st!=3)si.gov=gov;
    if(st!=3)si.syntaxNo=srel;
    stI.sums.push_back(si);
  }
  stI.wordNo++;
}
    
int
EmPronoun::
findBest(AnteInfos& ais,StoryInfo& stI,float& totProb)
{
  int bstnum =-1;
  double bstprb=-1;
  AnteInfos::iterator aisI;
  int i=0;
  for(aisI=ais.begin();aisI!=ais.end();aisI++){
    AnteInfo& ai=(*aisI);
    //cerr<<"\t"<<i<<" "<<Tree::gst->toString(ai.ante->wrd)
    //      <<" "<<aisI->prob<<endl;
    if(ai.prob>bstprb){
      if(ai.pro)totProb+=ai.prob;
      bstnum=i;
      bstprb=aisI->prob;
    }
    i++;
  }
  if(bstnum<0)return bstnum;
  AnteInfo& ai=ais[bstnum];
  if(ai.ante){
    int refnu=ai.ante->refNo;
    ai.pro->refNo=refnu;
  }
  return bstnum;
}

void
EmPronoun::
annotate(AnteInfos& ais,StoryInfo& stI,SummaryInfo& si)
{
  float totp=0;
  int bstnum=findBest(ais,stI,totp);
  if(bstnum<0)return;
  AnteInfo& bst=ais[bstnum];
  float thresh=4*10e-4;  
  /* "I"'s outside of quotes are made non-anaphoric */
  if(!bst.pro->quoteNo&&whPerson(bst.pro->wrd)==FIRST)return;
  else if(whPerson( bst.pro->wrd)==SECOND)return;
  else if(totp<thresh)return;
  else if(bst.ante==NULL)return;
  assert(bst.ante->tree);
  Tree* rntAnte=refTree(bst.ante->tree);
  /* catch odd cases like 
     (S (NP (DT The) (NNS citizens)) (ADVP (PRP themselves)) ...)
  */
  if(!rntAnte)rntAnte=bst.ante->tree->parent;
  assert(rntAnte);
  Tree* rntPro=bst.pro->tree->parent;
  assert(rntPro);
  assert(rntAnte);
  stIndex refnumA=rntAnte->refnum;
  Tree* hbante=rntAnte->htree;
  assert(hbante);
  string labs= Tree::gst->toString(hbante->label);
  if(labs=="PRP"||labs=="PRP$"){
    refnumA=hbante->refnum;
    if(refnumA==0){
    return;
    }
  }
  if(refnumA==0){
    refnumA=refnumCounter++;
    rntAnte->refnum=refnumA;
  }
  rntPro->refnum=refnumA;
}


void
EmPronoun::
score(AnteInfos& ais,StoryInfo& stI,SummaryInfo& si)
{
  stIndex pro=si.wrd;
  int anspos=checkPronounPosition(stI);
  if(anspos<0){
    return;
  }
  EvalDatum& ed=evalData[anspos];
  float totp;
  /* bstnum is the place of the highest prob ante */
  int bstnum=findBest(ais,stI,totp);
  if(bstnum<0){
    si.refNo=-2;
    pthread_mutex_lock(&reslock);
    results[1]++;
    if(ed.reftype>1)results[0]++;
    pthread_mutex_unlock(&reslock);
    if(ed.reftype>1){  //total correct is 0 2 8 9 11
      if(ISIT(pro))subResults[0]++;
      else subResults[9]++;
    }
    else{
      if(ISIT(pro))subResults[3]++;
      else subResults[10]++;
    }
    return;
  }
  /* bst is the highest prob ante */
  AnteInfo& bst=ais[bstnum];
  /* "I"'s outside of quotes are made non-anaphoric */
  if(!bst.pro->quoteNo&&whPerson(bst.pro->wrd)==FIRST){
    results[1]++;
    si.refNo=-2;
    if(ed.reftype>1){
      results[0]++;
      subResults[11]++;
    }
    else{
      subResults[12]++;
    }
    return;
  }
  float thresh=4*10e-4;  
  if(scoreExplitives && (ed.reftype>1||
			 (bst.ante==NULL) ||
			 (whPerson(bst.pro->wrd)==SECOND) ||
			 (totp<thresh)
			 )){
    pthread_mutex_lock(&reslock);
    results[1]++;
    if(ed.reftype>1 && (
			(bst.ante==NULL) ||
			(whPerson( bst.pro->wrd)==SECOND) ||
			(totp<thresh)
			)){
      si.refNo=-2;
      if(bst.ante==NULL){  //its correct: 0, 2
	subResults[0]++;
      }
      else if(whPerson( bst.pro->wrd)==SECOND){
	subResults[11]++;
      }
      else if(totp<thresh){
	if(!ISIT(bst.pro->wrd))subResults[1]++;
	else subResults[2]++;
      }
      results[0]++;
    }
    else if(ed.reftype<=1){
      if(bst.ante==NULL){
	si.refNo=-2;
	subResults[3]++;
      }
      else if(whPerson(bst.pro->wrd)==SECOND){
	si.refNo=-2;
	subResults[12]++;
      }
      else if(totp<thresh){
	si.refNo=-2;
	if(ISIT(bst.pro->wrd))	subResults[4]++; // its real 0 2 13
	else subResults[5]++;                    // its guessed 0 2 3 4
      }
    }
    else {
      if(!ISIT(bst.pro->wrd))      subResults[6]++;
      else subResults[13]++;
    }
    pthread_mutex_unlock(&reslock);
    return;
  }
    
  if(ed.reftype>1){
    return;
  }
      
  /*
  cerr<<"PRO2 "<<ed.sentenceNo<<" "<<ed.wordNo<<" "<< ed.refnum<<" "<<bst.ante->sentNo<<" "<<bst.ante->wordNo<<" ";
  */
  int rn=ed.refnum;
  assert(rn>=0);
  bool scr=false;
  EvalData& eda=refnumToData[rn];
  assert(eda.size()>0);
  scr=finalScr(eda,bst,0);
  int arn=bst.ante->refNo;
  
  results[1]++;
  if(arn==rn){
    subResults[8]++;
    results[0]++;
  }
  else{
    subResults[7]++;
  }
  /*
  else {
    cerr<<"RIGHT ";
    subResults[8]++;
  }
  si.refNo=bst.ante->refNo;
  pthread_mutex_lock(&reslock);
  pthread_mutex_unlock(&reslock);
  */
}
  
int
EmPronoun::
finalScr(EvalData& eda, AnteInfo& bst,int useStart)
{
  int scr=0;
  EvalData::iterator edI=eda.begin();
  int sz=eda.size();
  int pos=0;
  for(;edI!=eda.end();edI++){
    EvalDatum& ed2=*edI;
    pos++;
    if(ed2.storyNo!=bst.ante->storyNo)continue;
    if(ed2.sentenceNo!=bst.ante->sentNo)continue;
    if((!useStart&&ed2.wordNo==bst.ante->wordNo)||
       (useStart&&ed2.pronNo==bst.ante->wordNo)){
      //cerr<<"Correct "<<ed2.sentenceNo<<" "<<ed2.wordNo<<endl;
      //cerr<<bst.sentNo<<" "<<bst.wordNo<<"\n";
      scr=1;
      break;
    }
  }
  return scr;
}

void
EmPronoun::
readAnswers()
{
  ifstream ifs("devAllPer.txt");
  //ifstream ifs("testAllPer.txt");
  assert(ifs);
  for(;;){
    EvalDatum ed(ifs);
    if(ed.storyNo<0)break;
    else evalData.push_back(ed);
    if(ed.refnum<0)continue;
    if(refnumToData.size()<=ed.refnum) refnumToData.resize(ed.refnum+1);
    refnumToData[ed.refnum].push_back(ed);
  }
}
    
EvalDatum::
EvalDatum(ifstream& ifs)
{
  int sn;
  ifs >> sn;
  if(!ifs){
    storyNo=-1;
    return;
  }
  storyNo=sn;
  ifs>>sentenceNo;
  ifs>>wordNo;
  ifs>>reftype;
  ifs>>refnum;
  ifs>>pronNo;
}

int
EmPronoun::
checkPronounPosition(StoryInfo& stI)
{
  int sz=evalData.size();
  int datumNo=0;
  for(;datumNo<sz;datumNo++){
    assert(datumNo<evalData.size());
    EvalDatum& ed = evalData[datumNo];
    //cerr<<ed.storyNo<<" "<<ed.sentenceNo<<" " << ed.wordNo<<" " << ed.pronNo
    //<<" "<<ed.refnum<<endl;
    if(ed.storyNo>stI.storyNo)break;
    if(ed.storyNo<stI.storyNo) continue;
    if(ed.sentenceNo<stI.sentNo)continue;
    if(ed.sentenceNo!= stI.sentNo)return -1;
    if(ed.wordNo<stI.wordNo)continue;
    if(ed.wordNo!=stI.wordNo)return -1;
    return datumNo;
  }
  return -1;
  //cerr<<"Could not find pronoun "<<stI.storyNo<<" "<<stI.sentNo<< " "<<stI.wordNo<<endl;
  //assert(datumNo<sz);
}

SummaryInfo::
SummaryInfo(int stno,int senno,int wrdno,int synno,int quno,Tree* ant)
  : storyNo(stno),sentNo(senno),wordNo(wrdno),syntaxNo(synno),
    quoteNo(quno),refNo(0),wrd(ant->label),
    pretrm(ant->parent->label),gov(0),tree(ant){}

void
SummaryInfo::
write(ostream& ofs)
{
  ofs<<sentNo<<" "<<wordNo<<" "<<syntaxNo<<" ";
  ofs<<quoteNo<<" ";
  string wrds=Tree::gst->toString(wrd);
  ofs<<wrds<<" ";
  ofs<<Tree::gst->toString(pretrm)<<" ";
  if(gov>0)ofs<<Tree::gst->toString(gov);
  else ofs<<"^";
  ofs<<"\n";
}

void
SummaryInfo::
read(ifstream& ifs)
{
  ifs>>sentNo;
  ifs>>wordNo;
  ifs>>syntaxNo;
  ifs>>quoteNo;
  refNo=0;
  string tmp;
  ifs>>tmp;
  wrd=Tree::gst->toIndex(tmp);
  ifs>>tmp;
  pretrm=Tree::gst->toIndex(tmp);
  ifs>>tmp;
  if(tmp=="^")gov=0;
  else gov=Tree::gst->toIndex(tmp);
}

void
SummaryInfo::
read(istream& ifs)
{
  ifs>>sentNo;
  ifs>>wordNo;
  ifs>>syntaxNo;
  ifs>>quoteNo;
  refNo=0;
  string tmp;
  ifs>>tmp;
  wrd=Tree::gst->toIndex(tmp);
  ifs>>tmp;
  pretrm=Tree::gst->toIndex(tmp);
  ifs>>tmp;
  if(tmp=="^")gov=0;
  else gov=Tree::gst->toIndex(tmp);
}

double
EmPronoun::
pGenGHead(AnteInfo& ai)
{
  if(action==TRAIN&&stage<GEN)return 1;
  stIndex anteCmb =ai.ante->pretrm+(50*ai.ante->wrd);
  double prb=ai.prob;
  int pg;
  stIndex pro=ai.pro->wrd;
  int proNum=whPronoun(pro);
  if(proNum>=11&& proNum<=15)return 1;  //thirdperson plural does not indicate gender;
  stIndex proCmb =ai.pro->pretrm+(50*ai.pro->wrd);
  double ans=0;
  pthread_mutex_lock(&pgghlock);
  PGen& pG=genMap[anteCmb];
   if(action==TRAIN){
    if(whPerson(pro)==THIRD){
      pg=proGen(pro);
      ans=pG.prob[pg];
      if(prb>=0){
	PGen& pG2=genMap[ai.ante->pretrm];
	pG.pos[pg]+=prb;
	pG2.pos[pg]+=prb;
      }
    }
    else{
      int gen=0;
      double denom=0;
      if(prb>=0)for(gen=0;gen<2;gen++){
	denom+= pG.prob[gen];
      }
      //denom is to prorate the partial count according to the current prob;
      //I do not really understand what I am doing here.
      for(gen=0;gen<2;gen++){
	double ans1=pG.prob[gen];
	ans1*=genMap[proCmb].prob[gen];
	ans+=ans1;
	if(prb>0){
	  assert(denom>0);
	  double proratePrb=(pG.prob[gen]/denom)*prb;
	  pG.pos[gen]+=proratePrb;
	  genMap[proCmb].pos[gen]+=proratePrb;
	  genMap[ai.ante->pretrm].pos[gen]+=proratePrb;
	  genMap[ai.pro->pretrm].pos[gen]+=proratePrb;
	}
      }
    }
    pthread_mutex_unlock(&pgghlock);
    return ans;
  }

  GenMap::iterator gmI=genMap.find(anteCmb);
  if(whPerson(pro)!=THIRD){
    for(int gen=0;gen<2;gen++){
      double ans1=genMap[proCmb].prob[gen];
      if(gmI==genMap.end())
	ans1*=genMap[ai.ante->pretrm].prob[gen];
      else ans1*=gmI->second.prob[gen];
      ans+=ans1;
    }
    pthread_mutex_unlock(&pgghlock);
    return ans;
  }
  pg=proGen(pro);
  if(gmI== genMap.end()){
    ans=genMap[ai.ante->pretrm].prob[pg];
    pthread_mutex_unlock(&pgghlock);
    return ans;
  }
  else{
    ans=    gmI->second.prob[pg];
    pthread_mutex_unlock(&pgghlock);
    /*
    cerr<<"\t"
      <<Tree::gst->toString(ai.ante->wrd)<<" "
      <<ai.ante->pretrm<<" "
      <<pg<<" "<<ans<<endl;
    */
    return ans;
  }
}


void
EmPronoun::
setquote(Tree* t,StoryInfo& stI)
{
  Tree* p=t->parent;
  if(!p)return;
  string lab=Tree::gst->toString(p->label);
  if(lab=="``")stI.doubleQuote=stI.sentNo;
  if(lab=="''")stI.doubleQuote=-1;
}

void
EmPronoun::
procWh(int i,StoryInfo& stI)
{
  SummaryInfos& sums=stI.sums;
  SummaryInfo& si=sums[i];
  if(stage<GEN)return;
  int j=i-1;
  SummaryInfo* si2=NULL;
  for(;j>=0;j--){
    SummaryInfo& si3=sums[j];
    if(si3.wordNo==si.quoteNo){
      si2=&si3;
      break;
    }
  }
  assert(si2);
  int whw=whWh(si.wrd);
  stIndex anteCmb =si2->pretrm+(50*si2->wrd);
  pthread_mutex_lock(&pgghlock);
  PGen& pG=genMap[anteCmb];
  PGen& pG2=genMap[si2->pretrm];
  if(whw!=0){
    int pg=2;
    pG.pos[pg]+=1;
    pG2.pos[pg]+=1;
    pthread_mutex_unlock(&pgghlock);
    return;
  }
  double denom1=pG.prob[0]+pG.prob[1];
  if(denom1<=0){
    //cerr<<"WH "<<Tree::gst->toString(si2->wrd)<<" "
    //<<Tree::gst->toString(si.wrd)<<" "<<denom1<<" "<<pG.prob[2]
    //<<" "<<Tree::gst->toString(si2->pretrm)<<endl;
    pthread_mutex_unlock(&pgghlock);
    return;
  }
  double denom2=pG2.prob[0]+pG2.prob[1];
  assert(denom2>0);
  for(int gen=0;gen<2;gen++){
    pG.pos[gen]+=pG.prob[gen]/denom1;
    pG2.pos[gen]+=pG2.prob[gen]/denom2;
  }
  pthread_mutex_unlock(&pgghlock);
}


stIndex
EmPronoun::
npHead(Tree* t)
{
  if(sym(t)!="NP")return 0;
  Tree* ht =t->htree;
  if(sym(ht)!="POS")return ht->subtrees->label;
  Tree* p=t->subtrees;   
  for(;p;p=p->sibling){
    if(!p->sibling)      return p->htree->subtrees->label;
  }
  assert(false);
  return 0;
}

int
EmPronoun::
isBe(Tree* t)
{
  string bes[10]={"be","am","are","is","was","were","been","being","'s","'re"};
  string str=sym(t->subtrees);
  for(int i=0;i<10;i++)if(bes[i]==str)return 1;
  return 0;
}

int
isHave(Tree* t)
{
  string haves[4]={"have","has","had","having"};
  string str=sym(t->subtrees);
  for(int i=0;i<4;i++)if(haves[i]==str)return 1;
  return 0;
}

int
findSBAR(Tree* p)
{
  stIndex sb=Tree::gst->toIndex("SBAR");
  for(;p;p=p->sibling)    if(p->label==sb)return 1;
  return 0;
}

stIndex
selectAltGov(Tree* t)
{
  Tree* p=t->subtrees;
  int seenhead=0;
  for(;p;p=p->sibling){
    if(!p->sibling){
      return p->htree->subtrees->label;
    }
    if(p->htree==t->htree){
      seenhead=1;
      continue;
    }
    if(!seenhead)continue;
    stIndex lab=p->label;
    if(lab==Tree::gst->toIndex("ADVP"))continue;
    else if(lab==Tree::gst->toIndex("RB"))continue;
    else{
      return p->htree->subtrees->label;
    }
  }
  return 0;
}

stIndex
EmPronoun::
sGov(Tree* t,int& srel)
{
  Tree* p=t->subtrees;
  for(;p;p=p->sibling) if(sym(p)=="VP")return sGov(p,srel);
  Tree* ht=t->htree;
  int hasSBAR =findSBAR(t->subtrees);
  if(!isBe(ht)&&!isHave(ht)){
    if(sym(ht)=="VBN")srel=PASSOBJ;
    if(hasSBAR) srel=VERBSBAR;
    return ht->subtrees->label;
  }
  if(isBe(ht)){
    if(hasSBAR)srel=ISSBAR;
    else srel=IS;
  }
  else srel=HAVE;
  stIndex ans=selectAltGov(t);
  return ans;
}

stIndex
EmPronoun::
findGov(Tree* t,int& srel)
{
  /* t is pointing head word of np or posessive */
  Tree* ht=t->parent;
  Tree* p=ht->parent;
  srel=OTHER;
  assert(p); //p should be base np;
  if(sym(ht)=="PRP$"){
    stIndex ans=npHead(p);
    srel=POS;
    return ans;
  }
  p=p->parent;
  assert(p);
  string posstr=sym(p->htree);
  if(posstr=="POS"){
    srel=POS;
    //return posHead(p,0);
    return npHead(p);
  }
  for(;p;p=p->parent) if(p->htree!=ht)break;
  if(!p)return 0;
  string symp=sym(p);
  if(symp=="NP"){
    srel=NP;
    return p->htree->subtrees->label;
  }
  if(symp=="PP"){
    /* p is the pp node */
    stIndex hd=p->htree->subtrees->label;
    int majprep=whPrep(hd);
    if(majprep<0)return hd;
    srel=NP+1+majprep;
    Tree* nht=p->htree;
    /* nht is the IN node */
    for(;p;p=p->parent) if(p->htree!=nht)break;
    if(!p)return 0;
    else{
      return p->htree->subtrees->label;
    }
  }
  if(symp!="S"){

    if(symp[0]=='S')srel=SUBJECT;
    if(symp=="VP")srel=OBJECT;
    return p->htree->subtrees->label;
  }
  else{
    srel=SUBJECT;
    stIndex ans=sGov(p,srel);
    return ans;
  }
}

int
EmPronoun::
syntaxType(Tree* t)
{
  Tree* p=t->parent->parent;
  if(!p)return 2;
  for(p=p->parent;p;p=p->parent){
    string labstr=Tree::gst->toString(p->label);
    if(p->htree==t->parent){
      continue;
    }
    if(labstr[0]=='S')return 0;
    if(labstr=="VP")return 1;
    else return 2;
  }
  return 2;
}
