

#include "ECString.h"
#include "Tree.h"
#include <iostream>
#include "Gov3.h"
#include <set>
typedef vector<Tree*> Trees;
typedef vector<Trees> Words;
typedef map<stIndex,int> SiMap;

#define TRAIN 0
#define TEST  1
#define SUMMARY 2
#define TOTGOVS 3
#define SETDS 4
#define FIRST 0
#define SECOND 1
#define THIRD 2

#define ANTE2 1
#define ANTE3 2
#define GEN   3
#define GOV   4

#define SUBJECT 0
#define OBJECT 1
#define OTHER 2
#define WHSIGNAL 3
#define POS 4
#define IS 5
#define NP 6
#define PASSOBJ 28
#define HAVE 29
#define ISSBAR 30
#define VERBSBAR 31
#define LASTSYN 32
#define ISIT(x) (whPronoun(x)==8)

class DdTrip
{
 public:
  double prob;
  double pos;
  double cnt;
};

class SummaryInfo
{
 public:
  SummaryInfo(){}
  SummaryInfo(int stno,int senno,int wrdno,int synno,int quono,Tree* ant);
  void write(ostream& ofs);
  void read(ifstream& ifs);
  void read(istream& ifs);
  int storyNo;
  int sentNo;
  int wordNo;
  int syntaxNo;
  int quoteNo;
  int refNo;
  stIndex wrd;
  stIndex pretrm;
  stIndex gov;
  Tree* tree;
};

typedef vector<SummaryInfo> SummaryInfos;
typedef vector<DdTrip> DdTrips;
typedef pair<double,double> DdPair;

class AnteInfo
{
 public:
  AnteInfo(SummaryInfo* si1, SummaryInfo* si2) : pro(si1),ante(si2),prob(-1){}
  double prob;
  SummaryInfo* pro;
  SummaryInfo* ante;
};

typedef vector<AnteInfo> AnteInfos;

class EvalDatum
{
 public:
  EvalDatum() {}
  EvalDatum(ifstream&);
  int storyNo;
  int sentenceNo;
  int wordNo;
  int reftype;
  int refnum;
  int pronNo;
};

typedef vector<EvalDatum> EvalData;

class  PGen
{
 public:
  PGen() {for(int i=0;i<3;i++){prob[i]=1;pos[i]=0;}}
  double prob[3];
  double pos[3];
};

typedef map<stIndex,PGen> GenMap;

class StoryInfo
{
 public:
  StoryInfo() : storyNo(0),sentNo(0),wordNo(0),doubleQuote(-1){}
  int storyNo;
  int sentNo;
  int wordNo;
  int doubleQuote;
  SiMap refCount; 
  SummaryInfos sums;
};

class EmPronoun
{
 public:
  EmPronoun() : action(0),stage(0),iter(0),scoreExplitives(1),
    neglogp(0),refnumCounter(1){setPronouns();}
  void nextTree(Tree* t);
  void resetStory();
  void outputStory(ostream& os,StoryInfo& stI);
  void writeDictionary(ofstream& dictstr);
  void readDictionary(ifstream& dictstr);
  void summarize(Tree* t,StoryInfo& stI);
  void proc(int i,StoryInfo& sti);

  void mStep();
  void readAnswers();
  void readProbs();
  void writeProbs();
  static int syntaxConstraint(AnteInfo ai);
  static int whPronoun(stIndex t);
  static stIndex cannonize(stIndex t);
  static int whPrep(stIndex lab);
  static int whWh(stIndex t);
  static int whPerson(stIndex lab);
  static int whPossessivePronoun(stIndex t);
  static int whReflexivePronoun(stIndex wrd);
  int anteType(stIndex sti, stIndex wi,StoryInfo& stI);
  static int proType(stIndex wrd);
  static int isPlural(stIndex sti);
  static int proGen(stIndex sti);
  stIndex posHead(Tree* t,int posProp);
  stIndex npHead(Tree* t);
  int     isBe(Tree* t);
  stIndex sGov(Tree* t,int& srel);
  stIndex findGov(Tree* t,int& srel);
  void readJR();
  void nreadJR();
  int  results[2];
  int subResults[15];
  int  action;
  int  stage;
  int  iter;
  int  maxIter;
  int  scoreExplitives;
  float neglogp;
  Gov gov;
 private:
  void setPronouns();
  void setquote(Tree* t,StoryInfo& stI);
  double computeAnteProb(AnteInfo& ai,StoryInfo& stI);
  bool isNPHead(Tree* tr, Tree* wrd);
  double pAnteGDist(AnteInfo& ai,StoryInfo& stI);
  double pNumGNum(AnteInfo& ai);
  double pGenGHead(AnteInfo& ai);
  double pGovGAnte(AnteInfo& ai);
  double p12GHead(AnteInfo& ai);
  double pPersonGPerson(AnteInfo& ai);
  int finalScr(EvalData& eda, AnteInfo& bst,int useStart);
  int findBest(AnteInfos& ais,StoryInfo& stI,float& tp);
  void score(AnteInfos& ais,StoryInfo& stI,SummaryInfo& si);
  void annotate(AnteInfos& ais,StoryInfo& stI,SummaryInfo& si);
  void scoreJavaRap(EvalDatum& ed);
  void nscoreJavaRap(EvalDatum& ed);
  int bucketDist(int dist);
  int checkPronounPosition(StoryInfo& stI);
  int syntaxType(Tree* t);
  void procWh(int i,StoryInfo& sti);
  int  whAnte(Tree* t);
  double distributeProb(double ans,int cmb);
  int repeat(AnteInfo& ai);
  void readCommon();
  EvalData& findCorEd(EvalDatum& ed);
  void naddToOPClasses(EvalDatum& ed);
  stIndex refnumCounter;
  static vector<stIndex> Pronouns;
  static vector<stIndex> Whwords;
  static vector<stIndex> Preps;
  EvalData evalData;
  vector<EvalData> refnumToData;
  /* need protecting for threads */
  DdTrip  pNonAna;
  DdTrips pAntegDist;
  DdPair pNumgNum[3][2];
  DdPair pPersongPerson[12][3];
  GenMap genMap;
};
   

