#include "Tree.h"
#include <fstream>
#include <sys/resource.h>
#include <iostream>
#include <stdlib.h>
#include "headFinder.h"
#include "ntInfo.h"
#include <string.h>

vector<Tree*> Tree::story;
static pthread_mutex_t stlock = PTHREAD_MUTEX_INITIALIZER;
int Tree::inputNo =0;
int Tree::storyDone =0;

int Tree::singleLineReader=1; //set to 0 to read prettyprinted trees.
bool Tree::writeStrip=0;
SymbolTable* Tree::gst=NULL;
Tree::Tree(Tree& other):
	numprev(other.numprev),
	refnum(other.refnum),
	reftype(other.reftype),
	label(other.label),
	subtrees(NULL),
	sibling(NULL),
	parent(NULL)
{
	for(int j=0;j<10;j++)
	{
		vals[j] = other.vals[j];
	}

	if(other.subtrees)
	{
		subtrees = new Tree(*other.subtrees);

		for(Tree* st = subtrees;
			st != NULL;
			st = st->sibling)
		{
			st->parent = this;
		}
	}

	if(other.sibling)
	{
		sibling = new Tree(*other.sibling);
	}

	htree = hTreeFromTree();
}

void
Tree::
writeYield(ostream& os)
{
  if(!subtrees)os<<gst->toString(label)<<" ";
  else{
    Tree* p;
    for(p=subtrees;p;p=p->sibling) p->writeYield(os);
  }
}

Tree*
Tree::
hTreeFromTree()
{
  if(!subtrees)return this;
  if(!subtrees->subtrees)return this;
  else{
    int i,hp=headPosFromTree(this);
    Tree* p=subtrees;
    for(i=0;i<hp;i++)p=p->sibling;
    return p->hTreeFromTree();
  }
  return NULL;
}

void
Tree::
writeVals(ostream& os)
{
  int i;
  bool last = false;
  for(i=0;i<10;i++)
    if(vals[i]){
      if(last) os<<".";
      last=true;
      os<<i<<vals[i];
    }
}


void
Tree::
write(ostream& os, SymbolTable* st)
{
  if(!subtrees) os << word;
  else{
    string lab=st->toString(label);
    Tree* p;
    os<< "(" << lab;
    if(refnum>0)os<<"#"<<refnum;
    for(p=subtrees;p;p=p->sibling){
      os<<" ";
      p->write(os,st);
    }
    os<<")";
  }
}

void
skipspaces(istream& is,int& cp,string& buf)
{
  int c;
  for(;;){
    if(cp>=buf.size()){
      if(Tree::singleLineReader)return;
      buf="";
      cp=0;
      getline(is,buf);
      if(!is)return;
    }
    c=buf[cp++];
    if(!isspace(c))break;
  }
  cp--;
}


char* reftypes[5]={"OBJREF","ACTION","SYNTAX","NONLOCOBJ","ENV"};

int
reftypeNum(char *tmp)
{
  int i;
  for(i=0;i<5;i++)
    if(!strcmp(reftypes[i],tmp)) return i+1;
  //fprintf(stderr,"U %s\n",tmp);
  return 0;
}

void
readLabelStuff(Tree* t, char *s, int brkpt,int n)
{
  int k=0,i;
  char tmp[256];
  for(i=brkpt;i<n;i++){
    if(s[i]=='#'){
      int startnum;
      k=startnum=i+1;

      if(s[i+1]!='-'){
	for(k=i+1;k<n;k++){
	  if(s[k]=='-')break;
	  tmp[k-brkpt-1]=s[k];
	}
	tmp[k-brkpt-1]='\0';
	startnum=++k;
	//fprintf(stderr,"OB %s %i\n",tmp,n);
	//t->reftype=reftypeNum(tmp);
      }
      else{
	k++; 
	startnum++;
      }
      for(;k<n;k++){
	if(s[k]== '~')break;
	tmp[k-startnum]=s[k];
      }
      tmp[k-startnum]='\0';
      cerr<<"TMP " <<tmp<<endl;
      t->refnum=atoi(tmp);
      t->refnum++;  //prevent 0 refnum;
    }
    /*
    else if(s[i]=='~'){
      int starttildanum=i+1;
      for(i=starttildanum;i<n;i++) tmp[i-starttildanum]=s[i];
      tmp[i-starttildanum]='\0';
      int sti=atoi(tmp);
      t->numprev=sti;
      //fprintf(stderr,"ST %i %s \n",sti,s);
    }
    */
  }
}

void
readLabelOpen(Tree* t, char *s, int brkpt,int n)
{
  int k=0,i;
  char tmp[256];
  for(i=brkpt;i<n;i++){
    if(s[i]=='#'){
      int startnum;
      k=startnum=i+1;
      if(s[i+1]!='-'){
	for(k=i+1;k<n;k++){
	  if(s[k]=='-')break;
	  tmp[k-brkpt-1]=s[k];
	}
	tmp[k-brkpt-1]='\0';
	startnum=++k;
      }
      else{
	k++; 
	startnum++;
      }
      t->refnum=atoi(tmp);
      assert(t->refnum);
      //t->refnum++;  //prevent 0 refnum;
    }
  }
}

stIndex
readlabel(istream& is,Tree* t,int& cp,string& buf, SymbolTable* si)
{
  int	 c, brkpt=0,n=0, i=0;
  char s[256];
  for(c= buf[cp++] ;(cp<=buf.size() && !isspace(c) && c!='(' && c!=')');c=buf[cp++]){
    s[n++]=c;
  }
  cp--;
  brkpt=n;
  for (i=1; i<n-1; i++){		// don't look at 1st or last char 
    if (strchr(CATSEP,s[i])) {		// if s[i] is in CATSEP
      brkpt= i;
      break;
    }
  }
  //readLabelStuff(t,s,brkpt,n);
  readLabelOpen(t,s,brkpt,n);
  s[brkpt] = '\0';
  pthread_mutex_lock(&stlock);
  stIndex sti=si->toIndex(s);
  pthread_mutex_unlock(&stlock);
  t->label=sti;
}

stIndex
readGWord(istream& is,int& cp, string& buf, SymbolTable* si,string& S)
{
  char	 s[MAXLABELLEN];
  int	 n=0;
  char   c;
  for(c=buf[cp++];(cp<=buf.size() &&!isspace(c) && (c!='(') && (c!=')'));c=buf[cp++]){
    S.push_back(c);
    c=tolower(c);
    s[n++] = c;
    assert(n<MAXLABELLEN-1);		/* leave space for '\0' */
  }  
  s[n] = '\0';
  cp--;
  if(n==0)cerr<<"RGW "<<buf<<" "<<cp<<endl;
  if(n == 0) return 0;
  pthread_mutex_lock(&stlock);
  stIndex ans = si->toIndex((string)s);
  pthread_mutex_unlock(&stlock);
  //cerr<<"RGW "<< s<<" "<<ans<<endl;
  return ans;
}

void
addParent(Tree* tr,Tree* par)
{
  if(!tr)return;
  if(par) tr->parent=par;
  Tree* p;
  for(p=tr->subtrees;p;p=p->sibling)addParent(p,tr);
}

Tree*
Tree::
make(istream& is, SymbolTable* st)
{
  int cp=0;
  string buf;
  if(!singleLineReader)skipspaces(is,cp,buf);
  else{
    getline(is,buf);
  }
  if(!is) return NULL;
  if(buf.empty()&&singleLineReader){
    Tree* ns = new Tree(NEWSTORY);
    return ns;
  }
  Tree *ans=readB(is,cp,buf,st);
  addParent(ans,NULL);
  return ans;
}

Tree*
Tree::
readB(istream& is,int& cp, string& buf, SymbolTable* si)
{
  int 	c;
  Tree  *t,*p;
  skipspaces(is,cp,buf);

  c=buf[cp];
  cp++;
  switch (c) {

  case ')': 
    return(NULL); break;	/* empty tree */

  case '(':			/* nonterminal */
    skipspaces(is,cp,buf);
    t = new Tree(0);
    readlabel(is,t,cp,buf,si);
    for(;;){
      t->subtrees = p = readB(is,cp,buf,si);
      if(!terminal_p(t->label) && p &&!p->subtrees){
	delete p;
	t->subtrees=NULL;
      }
      else break;
    }
    while (p) {
      Tree* stree=readB(is,cp,buf, si);
      if(!terminal_p(t->label) && stree &&!stree->subtrees){
	delete stree;
      }
      else{
	p->sibling = stree;
	p = p->sibling;
      }
    }
    assert(t->label);
    t->htree=t->hTreeFromTree();
    return(t); break;
  
  default:		
    cp--;
    skipspaces(is,cp,buf);
    t = new Tree(0);
    t->sibling = t->subtrees=NULL;
    string S;
    t->label = readGWord(is,cp,buf,si,S);
    t->word=S;
    assert(t->label>0);
    return(t);
  }}


int
Tree::
hpos()
{
  if(!htree)return 0;
  Tree* p;
  int ans=0;
  for(p=subtrees;p;p=p->sibling){
    if(p->htree==htree)return ans;
    ans++;
  }
  cerr<<"No position for ";
  htree->write(cerr,gst);
  assert(0);
  return 0;
}
  
  

Tree*
Tree::
checkHTree()
{
  if(!htree)return NULL;
  Tree* p;
  for(p=subtrees;p;p=p->sibling){
    if(p->htree==htree)return p;
  }
  htree=hTreeFromTree();
  return htree;
}
  
int
Tree::
wPos()
{
  int pos=0;
  Tree* p;
  for(p=this;p->parent;p=p->parent){}
  if(p->wPos(this, pos))return pos;
  else return -1;
}
    
int
Tree::
wPos(Tree* ht,int& pos)
{
  Tree* p=subtrees;
  if(!p){
    pos++;
    return 0;
  }
  else if(this==ht) return 1;
  for(;p;p=p->sibling) if(p->wPos(ht,pos))return 1;
  return  0;
}
